/*
 * Copyright (C) 2005-2008  Equinox Software, Inc. / Georgia Public Library Service 
 * Mike Rylander <mrylander@gmail.com>
 * Copyright (C) 2010 Laurentian University
 * Dan Scott <dscott@laurentian.ca>
 * Copyright (C) 2013Equinox Software, Inc.
 * Mike Rylander <miker@esilibrary.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

DROP SCHEMA IF EXISTS actor CASCADE;

BEGIN;
CREATE SCHEMA actor;

CREATE TABLE actor.usr (
    id                  SERIAL  PRIMARY KEY,
    profile             INT     NOT NULL, -- user group
    email               TEXT    NOT NULL UNIQUE,
    passwd              TEXT    NOT NULL,
    photo_url           TEXT,
    prefix              TEXT,
    first_given_name    TEXT    NOT NULL,
    second_given_name   TEXT,
    family_name         TEXT    NOT NULL,
    suffix              TEXT,
    home_ou             INT     NOT NULL,
    active              BOOL    NOT NULL DEFAULT TRUE,
    super_user          BOOL    NOT NULL DEFAULT FALSE,
    barred              BOOL    NOT NULL DEFAULT FALSE,
    deleted             BOOL    NOT NULL DEFAULT FALSE
);

CREATE INDEX actor_usr_home_ou_idx ON actor.usr (home_ou);

CREATE FUNCTION actor.crypt_pw_insert () RETURNS TRIGGER AS $$
    BEGIN
        NEW.passwd = MD5( NEW.id || NEW.passwd );
        RETURN NEW;
    END;
$$ LANGUAGE PLPGSQL;

CREATE FUNCTION actor.crypt_pw_update () RETURNS TRIGGER AS $$
    BEGIN
        IF NEW.passwd <> OLD.passwd THEN
            NEW.passwd = MD5( NEW.id || NEW.passwd );
        END IF;
        RETURN NEW;
    END;
$$ LANGUAGE PLPGSQL;

CREATE TRIGGER actor_crypt_pw_update_trigger
    BEFORE UPDATE ON actor.usr FOR EACH ROW
    EXECUTE PROCEDURE actor.crypt_pw_update ();

CREATE TRIGGER actor_crypt_pw_insert_trigger
    BEFORE INSERT ON actor.usr FOR EACH ROW
    EXECUTE PROCEDURE actor.crypt_pw_insert ();

CREATE RULE protect_user_delete AS ON DELETE TO actor.usr DO INSTEAD UPDATE actor.usr SET deleted = TRUE WHERE OLD.id = actor.usr.id;

CREATE TABLE actor.usr_note (
    id          SERIAL  PRIMARY KEY,
    usr         INT     NOT NULL REFERENCES actor.usr ON DELETE CASCADE DEFERRABLE INITIALLY DEFERRED,
    creator     INT     NOT NULL REFERENCES actor.usr ON DELETE CASCADE DEFERRABLE INITIALLY DEFERRED,
    pub         BOOL    NOT NULL DEFAULT FALSE,
    title       TEXT    NOT NULL,
    value       TEXT    NOT NULL,
    create_date TIMESTAMP WITH TIME ZONE    DEFAULT NOW()
);
CREATE INDEX actor_usr_note_usr_idx ON actor.usr_note (usr);
CREATE INDEX actor_usr_note_creator_idx ON actor.usr_note ( creator );

CREATE TABLE actor.settings_group (
    name    TEXT PRIMARY KEY,
    label   TEXT UNIQUE NOT NULL -- I18N
);

CREATE TABLE actor.org_unit_setting_type (
    name            TEXT    PRIMARY KEY,
    label           TEXT    UNIQUE NOT NULL,
    grp             TEXT    REFERENCES actor.settings_group (name),
    description     TEXT,
    datatype        TEXT    NOT NULL DEFAULT 'string',
    fm_class        TEXT,
    view_perm       INT,
    update_perm     INT,
    --
    -- define valid datatypes
    --
    CONSTRAINT coust_valid_datatype CHECK ( datatype IN
    ( 'bool', 'integer', 'float', 'currency', 'interval',
      'date', 'string', 'object', 'array', 'link' ) ),
    --
    -- fm_class is meaningful only for 'link' datatype
    --
    CONSTRAINT coust_no_empty_link CHECK
    ( ( datatype =  'link' AND fm_class IS NOT NULL ) OR
      ( datatype <> 'link' AND fm_class IS NULL ) )
);

CREATE TABLE actor.usr_setting_type (
    name        TEXT PRIMARY KEY,
    label       TEXT UNIQUE NOT NULL,
    description TEXT,
    grp         TEXT    REFERENCES actor.settings_group (name),
    datatype    TEXT NOT NULL DEFAULT 'string',
    fm_class    TEXT,

    --
    -- define valid datatypes
    --
    CONSTRAINT coust_valid_datatype CHECK ( datatype IN
    ( 'bool', 'integer', 'float', 'currency', 'interval',
        'date', 'string', 'object', 'array', 'link' ) ),

    --
    -- fm_class is meaningful only for 'link' datatype
    --
    CONSTRAINT coust_no_empty_link CHECK
    ( ( datatype = 'link' AND fm_class IS NOT NULL ) OR
        ( datatype <> 'link' AND fm_class IS NULL ) )

);

CREATE TABLE actor.usr_setting (
    id      SERIAL  PRIMARY KEY,
    usr     INT     NOT NULL REFERENCES actor.usr ON DELETE CASCADE DEFERRABLE INITIALLY DEFERRED,
    name    TEXT    NOT NULL REFERENCES actor.usr_setting_type (name) ON DELETE CASCADE ON UPDATE CASCADE DEFERRABLE INITIALLY DEFERRED,
    value   TEXT    NOT NULL,
    CONSTRAINT usr_once_per_key UNIQUE (usr,name)
);
CREATE INDEX actor_usr_setting_usr_idx ON actor.usr_setting (usr);

CREATE TABLE actor.org_unit (
    id          SERIAL  PRIMARY KEY,
    parent_ou   INT     REFERENCES actor.org_unit (id) DEFERRABLE INITIALLY DEFERRED,
    owner       INT     REFERENCES actor.usr (id) DEFERRABLE INITIALLY DEFERRED,
    shortname   TEXT    NOT NULL UNIQUE,
    name        TEXT    NOT NULL UNIQUE,
    email       TEXT,
    phone       TEXT
);
CREATE INDEX actor_org_unit_parent_ou_idx ON actor.org_unit (parent_ou);

CREATE VIEW actor.org_unit_depth (id, depth) AS
WITH RECURSIVE t(id,depth) AS (
    SELECT id, 0 FROM actor.org_unit WHERE parent_ou IS NULL 
        UNION ALL
    SELECT a.id, depth + 1 FROM actor.org_unit a JOIN t ON (t.id = a.parent_ou)
) SELECT * FROM t;

CREATE OR REPLACE FUNCTION actor.org_unit_parent_protect () RETURNS TRIGGER AS $$
    DECLARE
        current_aou actor.org_unit%ROWTYPE;
        seen_ous    INT[];
        depth_count INT;
    BEGIN
        current_aou := NEW;
        depth_count := 0;
        seen_ous := ARRAY[NEW.id];

        IF (TG_OP = 'UPDATE') THEN
            IF (NEW.parent_ou IS NOT DISTINCT FROM OLD.parent_ou) THEN
                RETURN NEW; -- Doing an UPDATE with no change, just return it
            END IF;
        END IF;

        LOOP
            IF current_aou.parent_ou IS NULL THEN -- Top of the org tree?
                RETURN NEW; -- No loop. Carry on.
            END IF;
            IF current_aou.parent_ou = ANY(seen_ous) THEN -- Parent is one we have seen?
                RAISE 'OU LOOP: Saw % twice', current_aou.parent_ou; -- LOOP! ABORT!
            END IF;
            -- Get the next one!
            SELECT INTO current_aou * FROM actor.org_unit WHERE id = current_aou.parent_ou;
            seen_ous := seen_ous || current_aou.id;
            depth_count := depth_count + 1;
            IF depth_count = 100 THEN
                RAISE 'OU CHECK TOO DEEP';
            END IF;
        END LOOP;

        RETURN NEW;
    END;
$$ LANGUAGE PLPGSQL;

CREATE TRIGGER actor_org_unit_parent_protect_trigger
    BEFORE INSERT OR UPDATE ON actor.org_unit FOR EACH ROW
    EXECUTE PROCEDURE actor.org_unit_parent_protect ();

CREATE TABLE actor.org_unit_setting (
    id        BIGSERIAL    PRIMARY KEY,
    org_unit    INT        NOT NULL REFERENCES actor.org_unit ON DELETE CASCADE DEFERRABLE INITIALLY DEFERRED,
    name        TEXT    NOT NULL REFERENCES actor.org_unit_setting_type DEFERRABLE INITIALLY DEFERRED,
    value        TEXT        NOT NULL,
    CONSTRAINT ou_once_per_key UNIQUE (org_unit,name),
    CONSTRAINT aous_must_be_json CHECK ( sharestuff.is_json(value) )
);
CREATE INDEX actor_org_unit_setting_usr_idx ON actor.org_unit_setting (org_unit);

CREATE TABLE actor.usr_activity (
    id          BIGSERIAL   PRIMARY KEY,
    usr         INT         REFERENCES actor.usr (id) ON DELETE SET NULL,
    etype       INT         NOT NULL REFERENCES config.usr_activity_type (id),
    event_time  TIMESTAMPTZ NOT NULL DEFAULT NOW()
);

-- remove transient activity entries on insert of new entries
CREATE OR REPLACE FUNCTION actor.usr_activity_transient_trg () RETURNS TRIGGER AS $$
BEGIN
    DELETE FROM actor.usr_activity act USING config.usr_activity_type atype
        WHERE atype.transient AND
            NEW.etype = atype.id AND
            act.etype = atype.id AND
            act.usr = NEW.usr;
    RETURN NEW;
END;
$$ LANGUAGE PLPGSQL;

CREATE TRIGGER remove_transient_usr_activity
    BEFORE INSERT ON actor.usr_activity
    FOR EACH ROW EXECUTE PROCEDURE actor.usr_activity_transient_trg();

-- given a set of activity criteria, find the most approprate activity type
CREATE OR REPLACE FUNCTION actor.usr_activity_get_type (
        ewho TEXT,
        ewhat TEXT,
        ehow TEXT
    ) RETURNS SETOF config.usr_activity_type AS $$
SELECT * FROM config.usr_activity_type
    WHERE
        enabled AND
        (ewho  IS NULL OR ewho  = $1) AND
        (ewhat IS NULL OR ewhat = $2) AND
        (ehow  IS NULL OR ehow  = $3)
    ORDER BY
        -- BOOL comparisons sort false to true
        COALESCE(ewho, '')  != COALESCE($1, ''),
        COALESCE(ewhat,'')  != COALESCE($2, ''),
        COALESCE(ehow, '')  != COALESCE($3, '')
    LIMIT 1;
$$ LANGUAGE SQL;

-- given a set of activity criteria, finds the best
-- activity type and inserts the activity entry
CREATE OR REPLACE FUNCTION actor.insert_usr_activity (
        usr INT,
        ewho TEXT,
        ewhat TEXT,
        ehow TEXT
    ) RETURNS SETOF actor.usr_activity AS $$
DECLARE
    new_row actor.usr_activity%ROWTYPE;
BEGIN
    SELECT id INTO new_row.etype FROM actor.usr_activity_get_type(ewho, ewhat, ehow);
    IF FOUND THEN
        new_row.usr := usr;
        INSERT INTO actor.usr_activity (usr, etype)
            VALUES (usr, new_row.etype)
            RETURNING * INTO new_row;
        RETURN NEXT new_row;
    END IF;
END;
$$ LANGUAGE plpgsql;

COMMIT;
